package cn.edu.sdu.virion.feature.extraction;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import weka.core.Utils;
import cn.edu.sdu.virion.common.Constants;
import cn.edu.sdu.virion.common.Util;
import cn.edu.sdu.virion.core.Sequence;
import cn.edu.sdu.virion.support.database.DBHelper;
import cn.edu.sdu.virion.support.weka.Classify;

/**
 * ȡ-CTD
 * @author Administrator
 *
 */
public class CTD {
	
	public static String[] getClassifierClassNames()
	{
		return Constants.WEKA_CLASSIFIERS_CLASS_NAMES;
	}
	
	@SuppressWarnings("unchecked")
	public static void optimizeGroupSelection() throws Exception {
		int seqclassIndex = 2;
		int attributeStartIndex = 13;
		for (int i = 0; i < Constants.COMPOSITION_GROUP_COLLECTION.length; i++) {
			String selectTableName = "CDHIT_SEQS_" + (i + 1);
			String insertTableName = "ATTRS_EVAL";
			int groupCount = Constants.COMPOSITION_GROUP_COLLECTION[i].length;
			int featuresNum = groupCount + groupCount * (groupCount - 1) / 2
					+ groupCount * Constants.CTD_DISTRIBUTION_POSITIONS.length;
			int attributeEndIndex = attributeStartIndex + featuresNum - 1;
			String trainSql = String.format(
					"SELECT * FROM %s WHERE STATUS=1 ORDER BY SEQCLASS DESC",
					selectTableName);
			for (String classifierClassName : getClassifierClassNames()) {
				for (int j = 5; j <= 50; j+=5) {
					for (int k = 5; k <= 50; k+=5) {
						String[] removeOptions = Classify.prepareRemoveOptions(seqclassIndex,
								attributeStartIndex, attributeEndIndex);
						String[] classifierOptions = new String[6];
						classifierOptions[0] = "-I";
						classifierOptions[1] = String.valueOf(1 * j);
						classifierOptions[2] = "-K";
						classifierOptions[3] = String.valueOf(1 * k);;
						classifierOptions[4] = "-S";
						classifierOptions[5] = "1";
						Map result = Classify.crossValidate(trainSql, removeOptions.clone(),
								classifierClassName, classifierOptions.clone());
						result.put("INFO", String.format(
								"CTD-GoupIndex:%d-Classfier:%s-numFeatures:%s-numTrees:%s", i,
								classifierClassName, classifierOptions[3], classifierOptions[1]));
						String insertSql = DBHelper.prepareInsertSql(
								insertTableName, result);
						DBHelper.executeNonQuery(insertSql);
						System.out.println(insertSql);
					}
				}
			}

		}
	}

	
	public static void updateAll()
	{
		for (int i = 0; i < 7; i++) {
			Constants.CTD_COMPOSITION_GROUP = Constants.COMPOSITION_GROUP_COLLECTION[i];
			Constants.CTD_COMPOSITION_GROUP_LEADER = Constants.COMPOSITION_GROUP_LEADER_COLLECTION[i];
			updateComposition("CDHIT_SEQS_" + (i+1));
			updateTransition("CDHIT_SEQS_" + (i+1));
			updateDistribution("CDHIT_SEQS_" + (i+1));
		}
	}
	
	public static void updateComposition(String tableName)
	{
		updateField("Composition", tableName);
	}
	public static void updateDistribution(String tableName)
	{
		updateField("Distribution", tableName);
	}
	public static void updateTransition(String tableName)
	{
		updateField("Transition", tableName);
	}
	
	public static void updateWebSeqs(String jobid) {
		Constants.CTD_COMPOSITION_GROUP = Constants.COMPOSITION_GROUP_COLLECTION[5];
		Constants.CTD_COMPOSITION_GROUP_LEADER = Constants.COMPOSITION_GROUP_LEADER_COLLECTION[5];
		
		String tableName = "WEB_SEQS";
		String strWhere = String.format("JOBID='%s'", jobid);
		
		updateField("Composition", tableName, strWhere);
		updateField("Distribution", tableName, strWhere);
		updateField("Transition", tableName, strWhere);
	}
	
	private static void updateField(String fieldName, String tableName, String strWhere)
	{
		String sql = String.format("SELECT * FROM %s", tableName);
		if (strWhere != null && strWhere.length()>0) {
			sql += " WHERE " + strWhere;
		}
        Object[] params = new Object[]{};
        ResultSet rs = DBHelper.executeQuery(sql, params);
        try {
			while(rs.next()){
				Integer id = rs.getInt("ID");
				String sequence = rs.getString("SEQUENCE");
				float[] features = null;
				if (fieldName.equals("Composition")) 
					features = orderedComposition(sequence);
				else if (fieldName.equals("Transition")) 
					features = orderedTransition(sequence);
				else if (fieldName.equals("Distribution")) 
					features = orderedDistribution(sequence);
				String valueToSet = Utils.arrayToString(features);
				StringBuilder sb = new StringBuilder();
				for (int i = 1; i <= features.length; i++) 
					sb.append(String.format(",%s_%d='%f'", fieldName, i, features[i-1]));
				String insertSql = String.format(
						"UPDATE %s SET status=status%s WHERE ID=%d",
						tableName, sb.toString(), id);
				Object[] insertParams = new Object[]{};
				DBHelper.executeNonQuery(insertSql, insertParams);
			}
		} catch (SQLException e) {
			e.printStackTrace();
		}
        DBHelper.free(rs);
	}
	
	
	private static void updateField(String fieldName, String tableName)
	{	
		updateField(fieldName, tableName, null);
	}
	
	/**
	 * Composition
	 * @param sequence 
	 * @return Map
	 */
	public static Map Composition(Sequence sequence)
	{
		Map<Character,Float> resultMap = new HashMap<Character,Float>();
		Map<Character,Float> frequencyMap = sequence.getAcidsFrequency();
		for (int i = 0; i < Constants.CTD_COMPOSITION_GROUP_LEADER.length(); i++) {
			float sum = 0;
			for (Character c:Constants.CTD_COMPOSITION_GROUP[i].toCharArray()) {
				sum += frequencyMap.get(c);
			}
			resultMap.put(Constants.CTD_COMPOSITION_GROUP_LEADER.charAt(i), sum);
		}
		return resultMap;
	}
	
	public static float[] orderedComposition(String sequence)
	{
		Sequence seq = new Sequence(sequence);
		Map<Character,Float> map = Composition(seq);
		float[] result = new float[map.size()];
		for (int i = 0; i < Constants.CTD_COMPOSITION_GROUP_LEADER.length(); i++) {
			result[i] = map.get(Constants.CTD_COMPOSITION_GROUP_LEADER.charAt(i));
		}
		return result;
	}

	
	/**
	 * Distribution
	 * @param sequence 
	 * @return Map
	 */
	public static Map<String, Float> Distribution(Sequence sequence)
	{
		Map<String, Float> distributionMap = new HashMap<String, Float>();
		String seq = sequence.transform(Constants.CTD_COMPOSITION_GROUP, Constants.CTD_COMPOSITION_GROUP_LEADER);
		Map<Character,ArrayList<Integer>> recordMap = new HashMap<Character,ArrayList<Integer>>();
		for (int i = 0; i < Constants.CTD_COMPOSITION_GROUP_LEADER.length(); i++) {
			recordMap.put(Constants.CTD_COMPOSITION_GROUP_LEADER.charAt(i), new ArrayList<Integer>());
		}
		for (int n = 0; n < seq.length(); n++) {
			recordMap.get(seq.charAt(n)).add(n);
		}
		for (int i = 0; i < Constants.CTD_COMPOSITION_GROUP_LEADER.length(); i++) {
			ArrayList<Integer> positionRecords = recordMap.get(Constants.CTD_COMPOSITION_GROUP_LEADER.charAt(i));
			int countNum = positionRecords.size();
			for (float pos:Constants.CTD_DISTRIBUTION_POSITIONS) {
				int index = (int) (pos * countNum);
				index = index < 1 ? 1:index;
				distributionMap.put(
						String.format("%c at %f", Constants.CTD_COMPOSITION_GROUP_LEADER.charAt(i), pos),
						(float)index/seq.length());	
			}
		}
		return distributionMap;
	}
	
	public static float[] orderedDistribution(String sequence)
	{
		Sequence seq = new Sequence(sequence);
		Map<String,Float> map = Distribution(seq);
		float[] result = new float[map.size()];
		String groupLeader = Constants.CTD_COMPOSITION_GROUP_LEADER;
		int groupCount = groupLeader.length();
		int resultIndex = 0;
		for (int i = 0; i < groupCount; i++) {
			for (float pos:Constants.CTD_DISTRIBUTION_POSITIONS) {
				result[resultIndex] = map.get(String.format("%c at %f", groupLeader.charAt(i), pos));
				resultIndex++;
			}
		}
		return result;
	}
	
	
	
	/**
	 * Transition
	 * @param sequence 
	 * @return Map
	 */
	public static Map Transition(Sequence sequence)
	{	
		return Util.countToFrequency(TransitionMatchCount(sequence), sequence.getLength());
	}
	
	public static float[] orderedTransition(String sequence)
	{
		Sequence seq = new Sequence(sequence);
		Map<String,Float> map = Transition(seq);
		float[] result = new float[map.size()];
		String groupLeader = Constants.CTD_COMPOSITION_GROUP_LEADER;
		int groupCount = groupLeader.length();
		int resultIndex = 0;
		for (int k = 0; k < groupCount; k++)
			for(int m = k + 1; m < groupCount; m ++)
			{
				StringBuilder sb = new StringBuilder();
				sb.append(groupLeader.charAt(k)).append(groupLeader.charAt(m));
				result[resultIndex] = map.get(sb.toString());
				resultIndex++;
			}
		return result;
	}
	
	/**
	 * Transitionƥ
	 * @param sequence 
	 * @return Map
	 */
	private static Map TransitionMatchCount(Sequence sequence)
	{
		String seq = sequence.transform(Constants.CTD_COMPOSITION_GROUP, Constants.CTD_COMPOSITION_GROUP_LEADER);
		String groupLeader = Constants.CTD_COMPOSITION_GROUP_LEADER;
		int seqLen = seq.length();
		int groupCount = groupLeader.length();
		Map<String,Integer> countsMap = new HashMap<String,Integer>();
		for (int k = 0; k < groupCount; k++)
			for(int m = k + 1; m < groupCount; m ++)
			{
				char[] sub = {groupLeader.charAt(k), groupLeader.charAt(m)};
				String substr = String.valueOf(sub);
				countsMap.put(substr, 0);
				for (int n = 0; n < seqLen-1; n++)
					if ( (seq.charAt(n) == groupLeader.charAt(k) && seq.charAt(n+1) == groupLeader.charAt(m))
						 || (seq.charAt(n) == groupLeader.charAt(m) && seq.charAt(n+1) == groupLeader.charAt(k)) )
						countsMap.put(substr, countsMap.get(substr) + 1);
			}
		return countsMap;
	}
	
}
