package cn.edu.sdu.virion.support.weka;

import java.util.HashMap;
import java.util.Map;
import java.util.Random;


import weka.classifiers.Classifier;
import weka.classifiers.Evaluation;
import weka.classifiers.evaluation.ThresholdCurve;
import weka.core.Instances;
import weka.core.Utils;
import weka.experiment.InstanceQuery;
import weka.filters.Filter;
import weka.filters.unsupervised.attribute.Remove;
import weka.gui.visualize.PlotData2D;
import cn.edu.sdu.virion.common.Constants;
import cn.edu.sdu.virion.common.Util;

public class Classify {
		
	/**
	 * ݿ⣬ȡݼ
	 * @param sql ݿѯ
	 * @return
	 * @throws Exception 
	 */
	public static Instances queryInstances(String sql) throws Exception
	{
		InstanceQuery query = new InstanceQuery();
		query.setDatabaseURL(Constants.DB_URL);
		query.setUsername(Constants.DB_USER);
		query.setPassword(Constants.DB_PSWD);
		query.connectToDatabase();		
		query.setQuery(sql);
		Instances trainInitial = query.retrieveInstances();
		query.close();
		
		return trainInitial;
	}
	
	/**
	 * Ԥųݼе޹
	 * @param originInstancs ҪԤݼ
	 * @param removeOptions ų
	 * @return
	 * @throws Exception
	 */
	public static Instances preprocess(Instances originInstancs,
			String[] removeOptions) throws Exception {
		Remove remove = new Remove();
		remove.setOptions(removeOptions);
		remove.setInputFormat(originInstancs);
		Instances train = Filter.useFilter(originInstancs, remove);
		return train;
	}
	
	/**
	 * ֤
	 * @param trainSql ݼѯ
	 * @param removeOptions ųѡ
	 * @return
	 * @throws Exception
	 */
	public static Map<String, Float> crossValidate(String trainSql, String[] removeOptions, 
			String classifierName, String[] classifierOptions) throws Exception
	{
		/**
		 * ѵݼ
		 */
		Instances trainInitial = queryInstances(trainSql);
		
		/**
		 * Ԥųݼе޹
		 */
		Instances train = preprocess(trainInitial, removeOptions);
		train.setClassIndex(0);
		
		/**
		 * ʮ֤ý
		 */
		Evaluation eval = new Evaluation(train);
		eval.crossValidateModel(Classifier.forName(classifierName, classifierOptions), train, 10, new Random(1));
		double[][] confustionMatrix = eval.confusionMatrix();
		
		/**
		 * ROC
		 */
		ThresholdCurve tc = new ThresholdCurve();
		int classIndex = 0; // ROC for the 1st class label
		Instances curve = tc.getCurve(eval.predictions(), classIndex);
		PlotData2D plotdata = new PlotData2D(curve);
		plotdata.setPlotName(curve.relationName());
		plotdata.addInstanceNumberAttribute();
		System.out.println("Area under ROC = "
				+ Utils.doubleToString(ThresholdCurve.getROCArea(curve), 4));
// 		ThresholdVisualizePanel tvp = new ThresholdVisualizePanel();
//		tvp.setROCString("(Area under ROC = " +
//		Utils.doubleToString(ThresholdCurve.getROCArea(curve),4)+")");
//		tvp.setName(curve.relationName());
//		tvp.addPlot(plotdata);
		
		/**
		 * ָָ꣬Ϊ
		 */
		Map<String, Float> result = calculateNorm(confustionMatrix);
		result.put("ROC", (float) ThresholdCurve.getROCArea(curve));
		return result;
	}

	/**
	 * ָ
	 * @param confustionMatrix WEKA
	 * @return
	 */
	public static Map<String, Float> calculateNorm(double[][] confustionMatrix)
	{
		float TP = (float) confustionMatrix[0][0];
		float FN = (float) confustionMatrix[0][1];
		float FP = (float) confustionMatrix[1][0];
		float TN = (float) confustionMatrix[1][1];
		float Sensitivity = TP / (TP + FN);
		float Specificity = TN / (TN + FP);
		float Accuracy = (TP + TN) / (TP + FP + TN + FN);
		float MCC = (float) ((TP*TN - FP*FN) / Math.sqrt((TP+FN)*(TP+FP)*(TN+FP)*(TN+FN)));
		
		Map<String, Float> result = new HashMap<String, Float>();
		result.put("TP", TP);
		result.put("FN", FN);
		result.put("FP", FP);
		result.put("TN", TN);
		result.put("Sensitivity",Sensitivity);
		result.put("Specificity", Specificity);
		result.put("Accuracy", Accuracy);
		result.put("MCC", MCC);
		return result;
	}
	
	/**
	 * װݼԤųѡ
	 * @param seqclassIndex ֵ
	 * @param attributeStartIndex ʼֵ
	 * @param attributeEndIndex Խֵ
	 * @return
	 */
	public static String[] prepareRemoveOptions(int seqclassIndex, int attributeStartIndex, int attributeEndIndex)
	{
		String attributesIndice =  String.format("%d,%s", seqclassIndex,
				Utils.arrayToString(Util.range(attributeStartIndex, attributeEndIndex)));
		return prepareRemoveOptions(attributesIndice);
	}
	
	/**
	 * װݼԤųѡ
	 * @param seqclassIndex ֵ
	 * @param attributesIndice 
	 * @return
	 */
	public static String[] prepareRemoveOptions(int seqclassIndex, int[] attributesIndice)
	{
		return prepareRemoveOptions(String.format("%d,%s", seqclassIndex,
				Utils.arrayToString(attributesIndice)));
	}
	
	/**
	 * װݼԤųѡ
	 * @param attributesIndice
	 * @return
	 */
	public static String[] prepareRemoveOptions(String attributesIndice)
	{
		String[] options = new String[3];
		options[0] = "-R";
		options[1] = attributesIndice;
		options[2] = "-V";
		return options;
	}
}
